/*
Name :  set.c
Zweck:  Virtuelle Quorenmenge berechnen
Autor:  Andreas Seifert
Datum:
Update:
*/
/****************************************************************************/

#include "graph.h"
#include "set.h"

/****************************************************************************/


/*
 * AU: Wahrscheinlichkeitsberechnung
 * ST: computeVQM()             Berechne die virtuelle Quorenmenge
 *     buildVQM()               Ersetze virtuelle Rechner durch ihre 
 *                              Repraesentanten
 *     createSuccessfulSons()   Erzeuge Kombinationen aus Soehnen, deren
 *                              Summe an Stimmen die Quorenanzahl des Vaters 
 *                              erreicht 
 *     createPermutations()     Erzeuge eine Kombination
 *     addPermutationToFather() fuege Permutation zum Vaterknoten
 *     checkPermutation()       pruefe Kombination
 *     initVotingStructure()    Vorbereitung der Voting-Struktur auf die
 *                              Berechnung der virtuellen Quorenmenge
 *     printVQM()               Ausgabe virtuelle Quorenmenge
 *     printSons()              Ausgabe der Soehne eines Knotens, die
 *                              mit ihren Stimmen das Quorum des
 *                              Vaterknotens erreichen
 */



/**************************************************************************/
/*                     Beginn der Funktionen                              */
/**************************************************************************/


Bool computeVQM (graphicRoot)
     GRAPHIC_NODE *graphicRoot; /* Wurzelknoten der Voting-Struktur */
/*
 * AU: Berechne die virtuelle Quorenmenge
 * SU: clean_VQM()
 *     clean_ListOfReferences()
 *     addObjectToRefList()
 *     initVotingStructure()
 *     createSuccessfulSons()
 *     buildVQM()
 */
{
  Bool         foundSet = False; /* zeigt die erfolgreiche Bildung einer
                                  * Menge an */
  int          i;

  clean_VQM(); /* VQM initialisieren */

  clean_SuccessfulSons(); /* Knoten der Voting-Struktur initialisieren
                           * und die Reste vorheriger Berechnungen
                           * entfernen */

  /* fuege die Wurzel in die Referenzliste ein 
   * ptrCurrRefList ist der Zeiger auf das soeben eingefuegte Element 
   * ptrListOfReference zeigt auf den Beginn der Referenzen
   * die Uebergabe von NULL als ersten Parameter signalisiert die 
   * erstmalige Erzeugung eines Referenzlistenelementes 
   */

  for (i = 0; i < data.numb_quor; i++) { /* Berechnung der VQM fuer jeden
                                          * Dienst */

      clean_ListOfReferences(); /* Referenzliste initiaisieren */

      /* Interpretation der Voting-Struktur beginnt in der Wurzel, sie ist
       * daher als erstes Element der Referenzliste hinzuzufuegen */
      ptrCurrRefList = (REF_SET_ELEMENT *)addObjectToRefList(NULL, graphicRoot);

      ptrListOfReference = ptrCurrRefList;

      /* initialisiere die Votingstruktur und sortiere die Nachfolger 
       * nach Stimmgewicht */
      initVotingStructure(i);

      /* veranlasst die Traversierung der Votingstruktur */
      /* Erzeuge fuer jeden Knoten der Voting-Struktur Kombinationen aus
       * Soehnen, deren Summe an Stimmen die Quorenanzahl des Vaters erreicht
       */
      createSuccessfulSons(graphicRoot, i); 

/*      printSons(); Ausgabe der Kombinationen falls gewuenscht */

      /* Erzeugung vorlaeufiger VQM indem von der Wurzel startend in jedem
       * Knoten Kombinationen aus erfolgversprechenden Soehnen in die VQM 
       * aufgenommen werden, sie damit expandieren und ihrerseits wieder 
       * durch deren Soehne ersetzt werden
       */
      if (!buildVQM(ptrCurrRefList->ptrRefItem->firstNodeSetHead[i], 
                    &foundSet, i)) return False; 

      /* Ausgabe der virtuellen Quorenmenge je nach Operation */
      printVQM(i);
      } 
  return True;
}





Bool buildVQM(memberOfSet, foundSet, service)
     GRAPHIC_SET     *memberOfSet;
     Bool            *foundSet;
     int             service; 
/*
 * AU: Ersetze virtuelle Rechner durch ihre Repraesentanten
 * SU: addObjectSetToRefList()
 *     getRefListElement()
 *     buildVQM()
 *     addSolutionToVQMSet()
 *     removeObjectSetFromRefList()
 */ 
{
  GRAPHIC_SET         *graphicObjSet;        /* Zeiger auf virtuelle Quorenmenge */
  GRAPHIC_SET_ELEMENT *graphicObjSetElement; /* Zeiger auf Element der Menge */
 
  /* Es gibt kein Element, dass expandiert werden koennte */
  if ((ptrCurrRefList->ptrRefItem->firstNodeSetHead[service] == NULL) &&
      (ptrCurrRefList->ptrRefItem->first_arc != NULL)) return False;

  /* Durchlaufe die referenzliste und ersetze Knoten durch die Menge ihrer
   * erfolgreichen Soehne */
  for (graphicObjSet = ptrCurrRefList->ptrRefItem->firstNodeSetHead[service];
       graphicObjSet != NULL;
       graphicObjSet = graphicObjSet->nextNodeSetHead) {

      /* fuege die Menge der Soehne der Referenzliste hinzu */
      addObjectSetToRefList(ptrCurrRefList, graphicObjSet);

      *foundSet = True;

      /* ersetze jedes Element jedes Element rekursiv durch seine Soehne */
      for (graphicObjSetElement = graphicObjSet->firstNodeSetElement;
           graphicObjSetElement != NULL;
           graphicObjSetElement = graphicObjSetElement->nextNodeSetElement) {

          ptrCurrRefList = (REF_SET_ELEMENT *)
                           getRefListElement(graphicObjSetElement->ptrSetItem);

          if (!buildVQM(graphicObjSet, foundSet, service)) return False;
          }
      }

  /* das Ende der Referenzliste wurde erreicht */
  if (ptrCurrRefList->nextRefSetElement == NULL) {

     if (*foundSet) { /* eine virtuelle Gruppe wurde gefunden */

        addSolutionToVQMSet(service);

        *foundSet = False;
        }
     /* nicht erfolgreich gebildete Mengen sind zu entfernen */
     removeObjectSetFromRefList(memberOfSet);
     }
  return True;
}




void createSuccessfulSons(ptrGraphicObject, service)
     GRAPHIC_NODE *ptrGraphicObject;
     int          service;
/*
 * AU: Erzeuge fuer jeden Knoten der Voting-Struktur Kombinationen aus
 *     Soehnen, deren Summe an Stimmen die Quorenanzahl des Vaters erreicht
 *     Benutzter Algorithmus ist Depth-First-Search
 * SU: createPermutations()
 *     createSuccessfulSons() 
 */
{     
  GRAPHIC_ARC *ptrNextObject;
 
  /* Erzeuge die Kombinationen erfolgreicher Nachfolgerknoten  
   * fuer den gegenwaertigen Knoten */
  createPermutations(ptrGraphicObject, ptrGraphicObject->first_arc, service); 

  /* Aufruf fuer die Nachfolger des Knotens */
  for (ptrNextObject = ptrGraphicObject->first_arc; ptrNextObject != NULL;
       ptrNextObject = ptrNextObject->next_arc)
      createSuccessfulSons(ptrNextObject->arc_to_object, service); 
}     




void createPermutations(ptrFatherGraphicObject, ptrArcObject, service)
     GRAPHIC_NODE *ptrFatherGraphicObject; /* Vaterknoten */
     GRAPHIC_ARC  *ptrArcObject; /* Zeiger Verbindung zu einem Rechner */
     int          service;       /* Dienst */
/*
 * AU: Erzeuge rekursiv alle moeglichen Kombination der Soehne des 
 *     Vaterknotens und fuege eine erfolgreiche Kombination
 *     dem Vaterknoten hinzu
 * SU: addPermutationToFather()
 *     createPermutations()
 */
{ 
  if (ptrArcObject == NULL) {

     /* pruefe ob gute Menge, d.h. ob die Summe der Stimmen der als True 
      * markierten Rechner groesser oder gleich dem Quorum des Vaterknotens
      * ist und ob diese Menge k-minimal ist */

     if (checkPermutation(ptrFatherGraphicObject, service)) 
        addPermutationToFather(ptrFatherGraphicObject, service);
     }
  else { 
       /* nimm einen weiteren Rechner in die Kombination auf */
       ptrArcObject->arc_to_object->graphObjStatus = visited;

       createPermutations(ptrFatherGraphicObject, 
                          ptrArcObject->next_arc, service);

       /* entferne einen Rechner aus der Kombination */
       ptrArcObject->arc_to_object->graphObjStatus = non_visited;

       createPermutations(ptrFatherGraphicObject, 
                          ptrArcObject->next_arc, service);     
       }
}



void addPermutationToFather(ptrFatherGraphicObject, service)
     GRAPHIC_NODE *ptrFatherGraphicObject; /* Vaterknoten */
     int          service;                 /* Dienst */
/*
 * AU: fuege eine Kombination aus Nachfolgern, deren Summe zum das 
 *     Quorum des Vaterknotens erreicht, dem Vaterknoten hinzu
 * SU: createHeadOfNodeSet()
 *     addNodeSetElement()
 */
{
  GRAPHIC_ARC *ptrNextObject;
  GRAPHIC_SET *ptrGraphicObjectSet;

  /* Kopfelement einer neuen Menge erzeugen */
  ptrGraphicObjectSet = 
     (GRAPHIC_SET *)createHeadOfNodeSet(ptrFatherGraphicObject, service);

  for (ptrNextObject = ptrFatherGraphicObject->first_arc; 
       ptrNextObject != NULL;
       ptrNextObject = ptrNextObject->next_arc) {

      /* Kombination aus Nachfolgern dem Vaterknoten hinzufuegen */
      if (ptrNextObject->arc_to_object->graphObjStatus == visited) 
         addNodeSetElement(ptrGraphicObjectSet, ptrNextObject->arc_to_object);
      }
}



Bool checkPermutation(ptrGraphicObject, service)
     GRAPHIC_NODE *ptrGraphicObject; /* Vaterknoten der Rechner der 
                                      * Kombination */
     int          service;           /* Dienst */
/*
 * AU: pruefe Kombination aus Rechnern, ob die Summe ihrer Stimmen
 *     die Quorenanzahl ihres Vaterknotens erreicht und falls ja, ob
 *     die Kombination k-minimal ist
 */
{
  GRAPHIC_ARC *ptrNextObject;

  int sumOfVotes = 0;
  int differ = 0;

  /* Besitzt der Rechner Nachfolger ? */
  if (ptrGraphicObject->first_arc == NULL) return False;

  for (ptrNextObject = ptrGraphicObject->first_arc; ptrNextObject != NULL;
       ptrNextObject = ptrNextObject->next_arc) {

      /* Die Stimmen jedes an der Kombination beteiligten Rechners traegt
       * zum Erreichen des Qiorums bei */
      if (ptrNextObject->arc_to_object->graphObjStatus == visited) 
         sumOfVotes = sumOfVotes + ptrNextObject->arc_to_object->vote; 
      }

  /* Berechne Differenz zum Quorum des Vaterknotens */  
  differ = sumOfVotes - ptrGraphicObject->quor[service];

  if (differ < 0) return False; /* Summe Stimmen erreicht Quorum nicht */

  /* Falls mehr Stimmen als benoetigt vorhanden sind muss geprueft werden,
   * ob einer der beteiligten Rechner aus der Kombination entfernt werden
   * kann und das Quorum immer noch erreicht wird. Ist dies der Fall, 
   * ist die k-Minimalitaet verletzt.
   */
  for (ptrNextObject = ptrGraphicObject->first_arc; ptrNextObject != NULL;
       ptrNextObject = ptrNextObject->next_arc) {

      if ((ptrNextObject->arc_to_object->graphObjStatus == visited) &&
          (ptrNextObject->arc_to_object->vote <= differ)) return False;
      }
  return True;
}


void initVotingStructure(service)
     int service;
/*
 * AU: Vorbereitung der Voting-Struktur auf die Berechnung 
 *     der virtuellen Quorenmenge
 * SU: sortArc()
 */
{
  GRAPHIC_NODE *ptrGraphicObjectNode;

  for (ptrGraphicObjectNode = ptrGraphicObjectList;
       ptrGraphicObjectNode != NULL;
       ptrGraphicObjectNode = ptrGraphicObjectNode->next) {

      /* sortArc(ptrGraphicObjectNode->first_arc, service); das Sortieren der
       * Nachfolgerknoten ist evtl ueberfluessig */
      ptrGraphicObjectNode->graphObjStatus = non_visited;
      }
}


void printVQM(service)
     int service; /* Dienst */
/*
 * AU: Ausgabe der virtuellen Quorenmenge
 */
{
 VQM_SET *ptrSet;             /* Zeiger auf virtuelle  Quorenmenge */
 VQM_SET_ELEMENT *ptrElement; /* Zeiger auf Element einer 
                               * virtuellen Gruppe */

 for (ptrSet  = ptrVQMSet[service]; ptrSet!= NULL;
      ptrSet = ptrSet->nextVQMSetHead) {
     fprintf(stderr, "\n");
     for (ptrElement = ptrSet->firstVQMSetElement; ptrElement != NULL;
          ptrElement = ptrElement->nextVQMSetElement) {
         fprintf(stderr, " %s ", ptrElement->identVQMSet);
         }
     fprintf(stderr, "Operation %d\n", service + 1);
     }

 
}

void printSons()
/*
 * AU: Ausgabe der Mengen an Nachfolgerknoten, die das Quorum ihres
 *     Vaterknotens erreichen
 */
{
 GRAPHIC_NODE        *zwischenObject;
 GRAPHIC_SET         *zwischenSet;
 GRAPHIC_SET_ELEMENT *zwischenSetElem;

 fprintf(stderr, "\n");

 zwischenObject = ptrGraphicObjectList;

 while (zwischenObject != NULL) {
        fprintf(stderr, "Rechner %s Name %s \n ",
               zwischenObject->ident, zwischenObject->name);

        zwischenSet = zwischenObject->firstNodeSetHead[0];
        while (zwischenSet != NULL) {
               zwischenSetElem = zwischenSet->firstNodeSetElement;
               fprintf(stderr, "neues Set");
               while (zwischenSetElem != NULL) {
                     fprintf(stderr, "vertr. %s ", 
                             zwischenSetElem->ptrSetItem->ident);
                     zwischenSetElem = zwischenSetElem->nextNodeSetElement;
                     }
               fprintf(stderr, "\n");
        zwischenSet = zwischenSet->nextNodeSetHead;
               }

        fprintf(stderr, "\n");
        zwischenObject = zwischenObject->next;
        }
}



/***************************************************************************/
/*                    Ende der Funktionen                                  */
/***************************************************************************/
